.wy-side-nav-search,
.wy-nav-top {
    background: #5A46BE;
}

.wy-grid-for-nav,
.wy-body-for-nav,
.wy-nav-side,
.wy-side-scroll,
.wy-menu,
.wy-menu-vertical {
    background-color: #FFFFFF;
}

.wy-menu-vertical a:hover {
    background-color: #d9d9d9;
}

.btn-link:visited,
.btn-link,
a:visited,
.a.reference.external,
.a.reference.internal,
.wy-menu-vertical a,
.wy-menu-vertical li,
.wy-menu-vertical ul,
.span.pre,
.sig-param,
.std.std-ref,


html[data-theme=light] {
    --pst-color-inline-code: rgb(199, 37, 78) !important;
}

.sig-name {
    font-size: 1.25rem;
}
